import argparse
import sys

from .create_command import CreateCommand
from .init_command import InitCommand
from .make_command import MakeCommand
from .sync_command import SyncCommand
from .rollback_command import RollbackCommand


def get_command(command):
    """
    Let's be dynamic
    :return: command
    """
    commands = {'create': CreateCommand, 'init': InitCommand, 'make': MakeCommand, 'sync': SyncCommand,
                'rollback': RollbackCommand}
    return commands.get(command, None)


def parse_args(args):
    """
    解析命令行参数
    :param args: command line arguments
    :return: arguments
    """
    parser = argparse.ArgumentParser(description="A simple database migration tool")

    subparsers = parser.add_subparsers(help="commands", dest="command")
    # create
    create_parser = subparsers.add_parser("create", help="Create a new project")
    create_parser.add_argument('dirname', action='store', default='.',
                               help="The directory to create your project. (default: %(default)s)")

    # init
    init_parser = subparsers.add_parser("init", help="Init your project")
    init_parser.add_argument("-e", "--env", action="store", default="dev",
                               choices=['dev', 'prod'],
                               help="the enviroment you want to control. (default: %(default)s)")

    # make
    make_parser = subparsers.add_parser("make", help="Make a new migration")
    make_parser.add_argument("description", action="store",
                             help="provide a description. (for example: create_user_table)")

    # sync
    sync_parser = subparsers.add_parser("sync", help="Synchronize migrations")
    sync_parser.add_argument("-e", "--env", action="store", default="dev",
                               choices=['dev', 'prod'],
                               help="the enviroment you want to control. (default: %(default)s)")

    # rollback
    rollback_parser = subparsers.add_parser("rollback", help="Rollback migration")
    rollback_parser.add_argument("-e", "--env", action="store", default="dev",
                               choices=['dev', 'prod'],
                               help="the enviroment you want to control. (default: %(default)s)")

    arguments = parser.parse_args(args)

    if arguments.command is None:
        print(parser.print_help())
        sys.exit(-1)

    return arguments


def main():
    """
    命令工具户主入口
    :return: None
    """
    arguments = parse_args(sys.argv[1:])

    Command = get_command(arguments.command)
    if Command is None:
        print("An error occurred", file=sys.stderr)
        sys.exit(-1)

    try:
        command = Command(arguments)
        command.run()
    except Exception as e:
        print("Error: ", sys.exc_info()[1])



import os
from configparser import ConfigParser
from tempfile import NamedTemporaryFile
import subprocess

from .command import Command

GET_LATEST_MIGRATION = "'SELECT name FROM chunyun_migrations ORDER BY ID DESC LIMIT 1'"
REMOVE_LATEST_MIGRATION = "'DELETE FROM chunyun_migrations WHERE name = $${0}$$'"


class RollbackCommand(Command):

    def get_latest_migration(self, option):
        os.environ['PGPASSWORD'] = option.get(self.args.env, "password")
        cmd = "psql -h {host} -p {port}  -U {user} {name} -t -c {sql}".format(
                            host=option.get(self.args.env, 'host'),
                            port=option.get(self.args.env, 'port'),
                            user=option.get(self.args.env, 'user'),
                            name=option.get(self.args.env, 'database'),
                            sql=GET_LATEST_MIGRATION)
        handle = os.popen(cmd)
        output = handle.read().strip()
        handle.close()
        return output

    def remove_migration_record(self, option, name):
        os.environ['PGPASSWORD'] = option.get(self.args.env, "password")
        cmd = "psql -h {host} -p {port}  -U {user} {name} -c {sql}".format(
                            host=option.get(self.args.env, 'host'),
                            port=option.get(self.args.env, 'port'),
                            user=option.get(self.args.env, 'user'),
                            name=option.get(self.args.env, 'database'),
                            sql=REMOVE_LATEST_MIGRATION.format(name))
        handle = os.popen(cmd)
        output = handle.read()
        handle.close()

    def sync_migration(self, option, sql):
        handle = NamedTemporaryFile(delete=False)
        os.chmod(handle.name, 0o777)
        handle.write(sql.encode("utf-8"))
        handle.close()
        os.environ['PGPASSWORD'] = option.get(self.args.env, "password")
        cmd = "psql -h {host} -p {port}  -U {user} {name} -f '{file}'".format(
                            host=option.get(self.args.env, 'host'),
                            port=option.get(self.args.env, 'port'),
                            user=option.get(self.args.env, 'user'),
                            name=option.get(self.args.env, 'database'),
                            file=str(handle.name))
        cmd_handle = os.popen(cmd)
        output = cmd_handle.read()
        cmd_handle.close()
        os.unlink(handle.name)

    def run(self):
        parser = ConfigParser()
        parser.read("config.ini")

        # 获取migrations表里最新同步文件
        latest_migration = self.get_latest_migration(parser)

        if os.path.exists(os.path.join("migrations", latest_migration)):
            with open(os.path.join("migrations", latest_migration)) as handle:
                sql = handle.read()
                _, sql = sql.split("-- @down")
                self.sync_migration(parser, sql)
                self.remove_migration_record(parser, latest_migration)
                print("回滚{}".format(latest_migration))
        else:
            raise Exception("{}文件不存在！".format(latest_migration))



from abc import ABCMeta, abstractclassmethod
import os


class Command(metaclass=ABCMeta):
    """
    抽象命令类
    """

    def __init__(self, args):
        """
        初始化命令类，并接收参数
        :param args: 命令参数
        :return: None
        """
        self.args = args

    @abstractclassmethod
    def run(self):
        """
        执行命令
        :return: None
        """
        pass





import os
import random

from .command import Command


SQL_TPL = """
-- @up
---------------------------------------------------------------------
-- Migration SQL that makes the change goes here.

-- @down
---------------------------------------------------------------------
-- SQL to undo the change goes here.

"""


class MakeCommand(Command):

    def run(self):
        description = self.args.description
        files = os.listdir('migrations')
        files.sort()
        if len(files) == 0:
            filename = "001_{}.sql".format(description)
        else:
            file = files[-1]
            number, _ = file.split("_", 1)
            number = int(number)
            number += 1
            filename = "{:03d}_{:s}_{:d}.sql".format(number, description, random.randint(100000, 999999))

        with open(os.path.join('migrations', filename), "w") as handle:
            handle.write(SQL_TPL)

        print("{} is created.".format(filename))




import os
from configparser import ConfigParser
from tempfile import NamedTemporaryFile
import subprocess

from .command import Command

GET_LATEST_MIGRATION = "'SELECT name FROM chunyun_migrations ORDER BY ID DESC LIMIT 1'"
NEW_MIGRATION = "'INSERT INTO chunyun_migrations(name) VALUES($${0}$$)'"


class SyncCommand(Command):

    def get_latest_migration(self, option):
        os.environ['PGPASSWORD'] = option.get(self.args.env, "password")
        cmd = "psql -h {host} -p {port}  -U {user} {name} -t -c {sql}".format(
                            host=option.get(self.args.env, 'host'),
                            port=option.get(self.args.env, 'port'),
                            user=option.get(self.args.env, 'user'),
                            name=option.get(self.args.env, 'database'),
                            sql=GET_LATEST_MIGRATION)
        handle = os.popen(cmd)
        output = handle.read().strip()
        handle.close()
        return output

    def insert_migration_record(self, option, name):
        os.environ['PGPASSWORD'] = option.get(self.args.env, "password")
        cmd = "psql -h {host} -p {port}  -U {user} {name} -c {sql}".format(
                            host=option.get(self.args.env, 'host'),
                            port=option.get(self.args.env, 'port'),
                            user=option.get(self.args.env, 'user'),
                            name=option.get(self.args.env, 'database'),
                            sql=NEW_MIGRATION.format(name))
        handle = os.popen(cmd)
        output = handle.read()
        handle.close()
        #print(output)

    def sync_migration(self, option, sql):
        handle = NamedTemporaryFile(delete=False)
        os.chmod(handle.name, 0o777)
        handle.write(sql.encode("utf-8"))
        handle.close()
        os.environ['PGPASSWORD'] = option.get(self.args.env, "password")
        cmd = "psql -h {host} -p {port}  -U {user} {name} -f '{file}'".format(
                            host=option.get(self.args.env, 'host'),
                            port=option.get(self.args.env, 'port'),
                            user=option.get(self.args.env, 'user'),
                            name=option.get(self.args.env, 'database'),
                            file=str(handle.name))
        cmd_handle = os.popen(cmd)
        output = cmd_handle.read()
        cmd_handle.close()
        os.unlink(handle.name)

    def get_migration_files(self):
        files = os.listdir('migrations')
        files.sort()
        return files

    def run(self):
        parser = ConfigParser()
        parser.read("config.ini")

        # 获取migrations表里最新同步文件
        latest_migration = self.get_latest_migration(parser)
        files = self.get_migration_files()

        idx = -1
        try:
            idx = files.index(latest_migration)
        except:
            pass

        for file in files[idx+1:]:
            with open(os.path.join("migrations", file)) as handle:
                sql = handle.read()
                sql, _ = sql.split("-- @down")
                self.sync_migration(parser, sql)
                self.insert_migration_record(parser, file)
                print("{}同步到数据库".format(file))



from configparser import ConfigParser
import os
import shutil

from .command import Command


SQL_TPL = """
-- @up
---------------------------------------------------------------------

{up}

-- @down
---------------------------------------------------------------------

"""
INIT_MIGRATION_TABLE = "'CREATE TABLE IF NOT EXISTS chunyun_migrations(id serial primary key, name varchar not null, created_at TIMESTAMPTZ default CURRENT_TIMESTAMP(0));'"
INIT_RECORD = "'INSERT INTO chunyun_migrations(name) VALUES($$001_init.sql$$)'";


class InitCommand(Command):

    def dump(self, option):
        os.environ['PGPASSWORD'] = option.get(self.args.env, "password")
        handle = os.popen("pg_dump -h {host} -p {port}  -U {user} -x -O -s {name}".format(
                            host=option.get(self.args.env, 'host'),
                            port=option.get(self.args.env, 'port'),
                            user=option.get(self.args.env, 'user'),
                            name=option.get(self.args.env, 'database')))
        output = handle.read()
        handle.close()
        return output

    def create_migration_table(self, option):
        os.environ['PGPASSWORD'] = option.get(self.args.env, "password")
        cmd = "psql -h {host} -p {port}  -U {user} {name} -c {sql}".format(
                            host=option.get(self.args.env, 'host'),
                            port=option.get(self.args.env, 'port'),
                            user=option.get(self.args.env, 'user'),
                            name=option.get(self.args.env, 'database'),
                            sql=INIT_MIGRATION_TABLE)
        handle = os.popen(cmd)
        output = handle.read()
        handle.close()
        print(output)

    def insert_init_record(self, option):
        os.environ['PGPASSWORD'] = option.get(self.args.env, "password")
        cmd = "psql -h {host} -p {port}  -U {user} {name} -c {sql}".format(
                            host=option.get(self.args.env, 'host'),
                            port=option.get(self.args.env, 'port'),
                            user=option.get(self.args.env, 'user'),
                            name=option.get(self.args.env, 'database'),
                            sql=INIT_RECORD)
        handle = os.popen(cmd)
        output = handle.read()
        handle.close()
        print(output)

    def run(self):
        # 同步数据库
        if not os.path.exists("config.ini"):
            raise Exception("你必须先创建项目")

        pg_dump = shutil.which('pg_dump')
        if pg_dump is None:
            raise Exception("确保系统具备pg_dump命令")

        psql = shutil.which('psql')
        if pg_dump is None:
            raise Exception("确保系统具备psql命令")

        parser = ConfigParser()
        parser.read("config.ini")
        dumps = self.dump(parser)
        # 插入migration表
        self.create_migration_table(parser)
        print("创建migrations表")
        # 导出当前数据库结构

        if not os.path.exists("migrations"):
            os.makedirs("migrations")

        content = SQL_TPL.format(up=dumps)
        with open("migrations/001_init.sql", "w") as handle:
            handle.write(content)

        self.insert_init_record(parser)



from setuptools import setup

setup(
    name="chunyun",
    version="0.8.1",
    description="A simple database migration tool",
    url="http://github.com/erhuabushuo/chunyun",
    author="疯人院主任",
    author_email="erhuabushuo@gmail.com",
    license="MIT",
    packages=['chunyun'],
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'chunyun = chunyun.command_line:main'
        ]
    },
    test_suite='nose.collector',
    tests_require=['nose'],
)



import os

from .command import Command

CONFIG_TPL = """
[dev]
host = localhost
port = 5432
database = test
user = postgres
password =

[prod]
host = localhost
port = 5432
database = test
user = postgres
password =
"""


class CreateCommand(Command):

    def run(self):
        # 初始化config.ini文件
        config_path = os.path.join(self.args.dirname, 'config.ini')
        print("Creating config file {}".format(config_path))
        if os.path.exists(config_path):
            raise Exception("已经在目录下初始化过！")

        # 创建migrations目录
        migrations_path = os.path.join(self.args.dirname, "migrations")
        print("Creating directory {}".format(migrations_path))
        os.makedirs(migrations_path)

        with open(config_path, "w") as handle:
            handle.write(CONFIG_TPL)

        print("成功创建项目，请配置config.ini文件")
